"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.OIDCMockProvider = void 0;
const events_1 = require("events");
const http_1 = require("http");
const crypto_1 = __importDefault(require("crypto"));
const util_1 = require("util");
const url_1 = require("url");
async function randomString(n, enc) {
    return (await (0, util_1.promisify)(crypto_1.default.randomBytes)(n)).toString(enc);
}
function toJSONtoUTF8toBase64Url(obj) {
    return Buffer.from(JSON.stringify(obj), 'utf8').toString('base64url');
}
class OIDCMockProvider {
    constructor(config) {
        this.state = new Map();
        this.httpServer = (0, http_1.createServer)((req, res) => void this.handleRequest(req, res));
        this.config = config;
        this.issuer = '';
        this.kid = '';
        this.keys = {};
    }
    async init() {
        this.httpServer.listen(this.config.port ?? 0, this.config.hostname);
        await (0, events_1.once)(this.httpServer, 'listening');
        const { port } = this.httpServer.address();
        this.issuer = `http://${this.config.hostname ?? 'localhost'}:${port}`;
        this.kid = await randomString(8, 'hex');
        this.keys = await (0, util_1.promisify)(crypto_1.default.generateKeyPair)('rsa', {
            modulusLength: 2048,
        });
        return this;
    }
    static async create(config) {
        return await new this(config).init();
    }
    async close() {
        this.httpServer.close();
        await (0, events_1.once)(this.httpServer, 'close');
    }
    async handleRequest(req, res) {
        res.setHeader('Content-Type', 'application/json');
        try {
            const url = new URL(req.url || '/', this.issuer);
            if (req.method === 'POST') {
                if (req.headers['content-type'] !== 'application/x-www-form-urlencoded') {
                    throw new Error('Only accepting application/x-www-form-urlencoded POST bodies');
                }
                let body = '';
                for await (const chunk of req.setEncoding('utf8')) {
                    body += chunk;
                }
                const parsed = new url_1.URLSearchParams(body);
                url.search = new url_1.URLSearchParams({
                    ...Object.fromEntries(url.searchParams),
                    ...Object.fromEntries(parsed),
                }).toString();
            }
            let result;
            await this.config.overrideRequestHandler?.(url.toString(), req, res);
            if (res.writableEnded)
                return;
            switch (url.pathname) {
                case '/.well-known/openid-configuration':
                    result = {
                        issuer: this.issuer,
                        token_endpoint: new URL('/token', this.issuer).toString(),
                        authorization_endpoint: new URL('/authorize', this.issuer).toString(),
                        jwks_uri: new URL('/jwks', this.issuer).toString(),
                        device_authorization_endpoint: new URL('/device', this.issuer).toString(),
                    };
                    break;
                case '/jwks':
                    result = {
                        keys: [
                            {
                                alg: 'RS256',
                                kid: this.kid,
                                ...this.keys.publicKey.export({ format: 'jwk' }),
                            },
                        ],
                    };
                    break;
                case '/authorize': {
                    const { client_id, scope, response_type, redirect_uri, code_challenge, code_challenge_method, state, } = Object.fromEntries(url.searchParams);
                    if (response_type !== 'code') {
                        throw new Error(`unknown response_type ${response_type}`);
                    }
                    const redirectTo = new URL(redirect_uri);
                    redirectTo.search = new url_1.URLSearchParams({
                        code: await this.storeForSingleRetrieval({
                            client_id,
                            scope,
                            code_challenge,
                            code_challenge_method,
                        }),
                        state,
                    }).toString();
                    res.statusCode = 307;
                    res.setHeader('Location', redirectTo.toString());
                    res.end(`Moved to ${redirectTo.toString()}`);
                    break;
                }
                case '/token': {
                    const { code, code_verifier, device_code } = Object.fromEntries(url.searchParams);
                    const { client_id, scope, code_challenge, code_challenge_method, isDeviceCode, } = this.retrieveFromStorage(device_code ?? code);
                    if (!isDeviceCode) {
                        if (code_challenge_method !== 'S256') {
                            throw new Error(`Unsupported code challenge method ${String(code_challenge_method)}`);
                        }
                        const expectedChallenge = crypto_1.default
                            .createHash('sha256')
                            .update(code_verifier)
                            .digest('hex');
                        const actualChallenge = Buffer.from(code_challenge, 'base64').toString('hex');
                        if (expectedChallenge !== actualChallenge) {
                            throw new Error('Challenge mismatch');
                        }
                    }
                    const { access_token, id_token, expires_in } = await this.issueToken({
                        client_id,
                        scope,
                    });
                    result = {
                        access_token: access_token,
                        id_token: id_token,
                        refresh_token: await this.storeForSingleRetrieval({
                            id_token,
                            access_token,
                        }),
                        token_type: 'Bearer',
                        expires_in,
                    };
                    break;
                }
                case '/device': {
                    const { client_id, scope } = Object.fromEntries(url.searchParams);
                    const device_code = await this.storeForSingleRetrieval({
                        client_id,
                        scope,
                        isDeviceCode: true,
                    });
                    result = {
                        device_code,
                        user_code: await randomString(8, 'hex'),
                        verification_uri: new URL('/verify-device', this.issuer).toString(),
                        expires_in: 3600,
                        interval: 1,
                    };
                    break;
                }
                case '/verify-device':
                    res.statusCode = 200;
                    result = { status: 'Verified!' };
                    break;
                default:
                    res.statusCode = 404;
                    result = { error: 'not found:' + url.pathname };
                    break;
            }
            res.end(JSON.stringify(result));
        }
        catch (err) {
            res.statusCode = 500;
            res.end(JSON.stringify({
                error: typeof err === 'object' && err && 'message' in err
                    ? err.message
                    : String(err),
            }));
        }
    }
    async issueToken(metadata) {
        const { expires_in, payload } = await this.config.getTokenPayload(metadata);
        const currentTimeInSeconds = Math.floor(Date.now() / 1000);
        const header = {
            alg: 'RS256',
            typ: 'JWT',
            kid: this.kid,
        };
        const fullPayload = {
            jti: await randomString(8, 'hex'),
            iat: currentTimeInSeconds,
            exp: currentTimeInSeconds + expires_in,
            client_id: metadata.client_id,
            scope: metadata.scope,
            iss: this.issuer,
            aud: metadata.client_id,
            ...payload,
        };
        const makeToken = (payload) => {
            const signedMaterial = toJSONtoUTF8toBase64Url(header) +
                '.' +
                toJSONtoUTF8toBase64Url(payload);
            const signature = crypto_1.default
                .createSign('RSA-SHA256')
                .update(signedMaterial)
                .sign(this.keys.privateKey, 'base64url');
            return `${signedMaterial}.${signature}`;
        };
        return {
            expires_in,
            access_token: makeToken(fullPayload),
            id_token: metadata.skipIdToken
                ? undefined
                : makeToken({ ...fullPayload, aud: metadata.client_id }),
        };
    }
    async storeForSingleRetrieval(payload) {
        const id = await randomString(12, 'base64');
        this.state.set(id, payload);
        return id;
    }
    retrieveFromStorage(id) {
        const entry = this.state.get(id);
        this.state.delete(id);
        return entry;
    }
}
exports.OIDCMockProvider = OIDCMockProvider;
//# sourceMappingURL=index.js.map